/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GLVIEWER_H
#define GLVIEWER_H

#pragma once

#include "SelectionsManager.h"
#include "PhongShader.h"
#include <QWidget>
#include <QGLViewer/qglviewer.h>
#include <QString>
#include "Mesh.h"
#include "Debug.h"
#include <QProgressDialog>

enum VISU_MODE {
    USER_FIELD = 0,
    SIMILARITY_FIELD = 1,
    NO_VISUMODE = 2
};

/**
 *  GLViewer class
 */
class GLViewer : public QGLViewer  {

  Q_OBJECT

public:
    inline GLViewer( QWidget *parent) : QGLViewer(parent){
        _widget = new SelectionWidget;
        _selectionsManager = new SelectionsManager;
        meshInitialized = false;
    }

    //***************************************************
    //                  mesh I/O
    //***************************************************
    /**
     * @fn open a mesh file
     * @param mesh filename
   */
    void open (const QString & filename);

    /**
     * @fn save a mesh .off
     * @param filename of the saved mesh
     */
    void saveMesh (const QString & filename);


    //***************************************************
    //                  selections
    //***************************************************

    /**
     * @fn save the current selections
     * @param name of the file
     */
    void saveSelections(const QString &filename);

    /**
     * @fn open the current selections
     * @param name of the file
     */
    void loadSelections(const QString &filename);

    /**
     * @brief setSelectionType
     * @param t patch, part or component
     */
    void setSelectionType(TYPE t);

    /**
     * @brief setEraserMode
     * @param b erase mode if true
     */
    void setEraserMode(bool b);

    /**
     * @brief setDeleteSelectionMode
     * @param b delete selections if true
     */
    void setDeleteSelectionMode(bool b);

    /**
     * @brief setDrawStrokeMode
     * @param b true to refine the selection
     */
    void setDrawStrokeMode(bool b);

    /**
     * @fn save timing
     * @param name of the file
     */
    void saveTimings(const QString &filename);

    //***************************************************
    //                  expansion
    //***************************************************
    /**
     * @fn open the SVM selection
     * @param name of the file
     * @brief debugging function to see the results from the svm program
     */
    void openSVMselection(const QString & filename);

    /**
     * @brief getSCRefHistogram
     * @return return the reference hist
     */
    QPixmap getSCRefHistogram();

    /**
     * @brief getSCCurHistogram
     * @param dist distance between the current histogram and the ref one
     * @return return the histogram under the mouse
     */
    QPixmap getSCCurHistogram(float & dist);

    //***************************************************
    //                  cameras & video
    //***************************************************
    /**
     * @fn set the video name
     * @param name of the videos
     */
    void setVideoName(const QString & filename);

    /**
     * @fn save the current Camera
     * @param name of the camera file
     */
    void saveCamera(const QString &filename);

    /**
     * @fn open the current Camera
     * @param name of the camera file
     */
    void openCamera(const QString &filename);

protected :
    /**
     * @brief init initialize the viewer
     */
    void init();

    /**
     * @brief draw draw fn
     */
    void draw ();

    /**
     * @brief helpString to fill the help (press h to see it)
     * @return string to return
     */
    QString helpString() const;

    //***************************************************
    //                  events
    //***************************************************

    /**
     * @brief event
     * @param event
     * @return
     */
    bool event(QEvent *event);

    /**
     * @brief keyPressEvent
     * @param e
     */
    void keyPressEvent (QKeyEvent * e);

    /**
     * @brief clicToPoint3D return true if there is a point under (currentX, currentY) pixel
     * @param currentX x value of the pixel
     * @param currentY y value of the pixel
     * @param p position of the point (if any) under the pixel
     * @return true if p exists
     */
    bool clicToPoint3D(int currentX, int currentY, Vec3Df & p);

    /**
     * @brief mousePressEvent
     * @param e
     */
    void mousePressEvent(QMouseEvent* e);

    /**
     * @brief mouseMoveEvent
     * @param e
     */
    void mouseMoveEvent(QMouseEvent* e);

    /**
     * @brief mouseReleaseEvent
     * @param e
     */
    void mouseReleaseEvent(QMouseEvent* e);

    /**
     * @brief wheelEvent
     */
    void wheelEvent(QWheelEvent *);


    //***************************************************
    //                  mesh I/O
    //***************************************************
    /**
     * @brief openOFF
     * @param filename
     * @param normWeight
     * @param mesh
     */
    void openOFF (const std::string filename, unsigned int normWeight, Mesh & mesh);

    /**
     * @brief openOBJ
     * @param filename
     * @param normWeight
     * @param mesh
     */
    void openOBJ (const std::string filename, unsigned int normWeight, Mesh & mesh);

    /**
     * @brief initTexture
     */
    void initTexture();

private:

    //***************************************************
    //                    Lights & Drawing Fns
    //***************************************************
    /**
     * @brief initLights
     */
    void initLights ();

    /**
     * @brief setDefaultMaterial
     */
    void setDefaultMaterial ();

    /**
     * @brief setShaderValues (diff, spec, shini)
     */
    void setShaderValues ( );

    /**
     * @brief drawMesh
     */
    void drawMesh ();

    /**
     * @brief drawMeshDebug draw the mesh in debug mode
     */
    void drawMeshDebug();

    Mesh mesh; /**< mesh */
    bool meshInitialized; /**< true if a mesh has been loaded*/

    SelectionsManager* _selectionsManager; /**< selection manager*/
    SelectionWidget* _widget; /**< widget */
    unsigned int vertexUnderMouse; /**< index of the vertex under the mouse */


    // for the video
    bool _beginvideo; /**< true if we start the video*/

    // for the visualization
    bool wireframe; /**< true if we want to display the wireframe*/

    PhongShader *phongShader; /**< shader */
    float diffuseRef; /**< diffuse value for the shader*/
    float specRef; /**< specular value for the shader*/
    float shininess; /**< shininess for the shader */

    bool debug; /**< true if we are in debug mode */
    VISU_MODE _debugVisuMode; /**< set to similarity_field, or user field, or none*/
    int debugTextureId; /**< index of the texture currently used*/
    GLuint ui_text_id[2]; /**< texture id*/

    QColor _strokeColor;/**< color of the stroke */

    QColor _patchColor; /**< color for patches */
    QColor _partColor; /**< color for parts  */
    QColor _ccColor; /**< color for components */
    QColor _bgColor; /**< color for background */
    QColor _csColor; /**< */
    QColor _negColor; /**< */

public slots:
    //***************************************************
    //                   selection SLOTS
    //***************************************************
    /**
     * @brief startSelection
     */
    void startSelection();

    /**
     * @brief updateSelection
     */
    void updateSelection();

    /**
     * @brief endSelection
     */
    void endSelection();

    /**
     * @brief setSelection (changed with the e key)
     * @param b
     */
    void setSelection(bool b);

    /**
     * @brief acceptSelections
     */
    void acceptSelections();

    /**
     * @brief reverseCurrentSelection
     */
    void reverseCurrentSelection(){
        _selectionsManager->reverseCurrentSelection();
        updateGL();
    }

    /**
     * @brief deleteAllSelections
     */
    void deleteAllSelections(){
        _widget->clear();
        _widget->setMode(0);
        _selectionsManager->deleteAll();
        updateGL();
    }

    /**
     * @brief selectionTypeImgNeedUpdate
     * to emit a signal when the type of selection change
     * in order to change the VisuDock accordingly
     */
    void selectionTypeImgNeedUpdate();

    //***************************************************
    //                   visu SLOTS
    //***************************************************
    /**
     * @brief updateStrokeColor
     * @param c new color to use
     */
    void updateStrokeColor(QColor c);

    /**
     * @brief updatePatchColor
     * @param c new color to use
     */
    void updatePatchColor(QColor c);

    /**
     * @brief updatePartColor
     * @param c new color to use
     */
    void updatePartColor(QColor c);

    /**
     * @brief updateccColor
     * @param c new color to use
     */
    void updateccColor(QColor c);

    /**
     * @brief updatebgColor
     * @param c new color to use
     */
    void updatebgColor(QColor c);

    /**
     * @brief changeVisuMode
     * @param c change the debug mode
     */
    void changeVisuMode(int c);

    /**
     * @brief changeDebugTexture change the texture used in the debug mode
     * @param val new texture to use
     */
    void changeDebugTexture(int val);

    //***************************************************
    //                  Expansion  SLOTS
    //***************************************************
    /**
     * @brief findSimilarSelections
     */
    void findSimilarSelections();

    /**
     * @brief thresholdMSChanged change the number of similar selections
     * @param val number of similar selections asked
     */
    void thresholdMSChanged(int val);

    /**
     * @brief changePatchSimAlgo change the algorithm for patch expansion
     * @param i
     */
    void changePatchSimAlgo(int i);

    /**
     * @brief saveSelHist save the histogram of the reference selection
     */
    void saveSelHist(){_selectionsManager->saveSelectionHist();}

signals:
    void debugMode(bool b);
    void signalTextToStatusBar(QString t);
    void selectionTypeChange(TYPE t);
    void updateScRefHistogram();
    void updateScCurHistogram();
    void resetStrokeToDraw();
};

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
#endif  // GLVIEWER_H
