/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <iostream>
#include <vector>
#include "Vec3D.h"

/**
 * @brief The Vertex class
 */
class Vertex {
public:
    inline Vertex () 
        : pos (Vec3Df (0.0,0.0,0.0)), normal (Vec3Df (0.0, 0.0, 1.0)) {}

    inline Vertex (const Vec3Df & pos) 
        : pos (pos), normal (Vec3Df (0.0, 0.0, 1.0)) {}

    inline Vertex (const Vec3Df & pos, const Vec3Df & normal) : pos (pos), normal (normal) {}

    inline Vertex (const Vertex & v) : pos (v.pos), normal (v.normal) , smoothNormal (v.smoothNormal)
      , faces(v.faces)
    {}
    inline virtual ~Vertex () {}

    /**
     * @brief operator =
     * @param vertex
     * @return this = vertex
     */
    inline Vertex & operator= (const Vertex & vertex) {
        pos = vertex.pos;
        normal = vertex.normal;
        smoothNormal = vertex.smoothNormal;
        faces.clear();
        faces = vertex.getFaces();
        return (*this);
    }

    //**************************************************
    //                  ACCESSORS
    //**************************************************

    /**
     * @brief getPos
     * @return the Vec3D pos of the vertex
     */
    inline const Vec3Df & getPos () const { return pos; }

    /**
     * @brief getNormal
     * @return the normal
     */
    inline const Vec3Df & getNormal () const { return normal; }  

    /**
     * @brief getSmoothNormal
     * @return the smoothed normal
     */
    inline const Vec3Df & getSmoothNormal () const { return smoothNormal; }

    /**
     * @brief setPos set the vertex position to newPos
     * @param newPos
     */
    inline void setPos (const Vec3Df & newPos) { pos = newPos; }

    /**
     * @brief setNormal set the vertex normal to newNormal
     * @param newNormal
     */
    inline void setNormal (const Vec3Df & newNormal) { normal = newNormal; }

    /**
     * @brief setSmoothNormal set the vertex smoothed normal to newNormal
     * @param newNormal
     */
    inline void setSmoothNormal (const Vec3Df & newNormal) { smoothNormal = newNormal; }

    /**
     * @brief operator ==
     * @param v
     * @return true if this == v (in both position and normal)
     */
    inline bool operator== (const Vertex & v) { return (v.pos == pos && v.normal == normal); }

    /**
     * @brief interpolate this = u*(1-alpha)+v*alpha (for both position and normal)
     * @param u
     * @param v
     * @param alpha
     */
    void interpolate (const Vertex & u, const Vertex & v, float alpha = 0.5);

    /**
     * @brief computeAveragePosAndRadius compute the average center of a list of vertex, and the max radius
     * @param vertices
     * @param center
     * @param radius
     */
    static void computeAveragePosAndRadius (std::vector<Vertex> & vertices, 
                                            Vec3Df & center, float & radius);

    /**
     * @brief scaleToUnitBox scale the vertices so that they fit in a unit box center in 0
     * @param vertices
     * @param center
     * @param scaleToUnitBox
     */
    static void scaleToUnitBox (std::vector<Vertex> & vertices, 
                                Vec3Df & center, float & scaleToUnitBox);

    /**
     * @brief normalizeNormals
     * @param vertices
     */
    static void normalizeNormals (std::vector<Vertex> & vertices);

    /**
     * @brief addFace add a face t
     * @param t
     */
    void addFace(int t){faces.push_back(t);}

    /**
     * @brief getFaces
     * @return all the faces that contains this vertex
     */
    const std::vector<unsigned int> & getFaces() const { return faces; }
    
private:
    Vec3Df pos; /**< position of the vertex */
    Vec3Df normal; /**< normal of the vertex */
    Vec3Df smoothNormal; /**< smoothed normal of the vertex */
    std::vector<unsigned int> faces; /**< faces that contains the vertex */
};

extern std::ostream & operator<< (std::ostream & output, const Vertex & v);

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
