/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef HISTOGRAM1D_H
#define HISTOGRAM1D_H

#include "Debug.h"
#include <iostream>
#include <climits>
#include <math.h>
#include <vector>
#include <cfloat>

#include <QColor>
#include <QPixmap>
#include <QPainter>
#include <assert.h>


template<typename T, typename histT>
/**
 * @brief The Histogram1D class
 */
class Histogram1D
{
public:
    /**
     * @brief round_W (usefull only for windows..
     * @param d
     * @return same as the normal round function
     */
    int round_W(double d)
    {
        return static_cast<int>(d + 0.5);
    }

    Histogram1D()
    {
        values_sum = (histT) 0.f;
        xbin_nb = 0;
        bounds[0] = (T) FLT_MAX;
        bounds[1] = (T) FLT_MIN;
    }

    Histogram1D( unsigned int x_bin_nb,
                 T  x_bin_min, T x_bin_max )
    {
        xbin_nb = x_bin_nb;
        values_sum = (histT) 0.f;
        hist.clear();
        for (unsigned int i = 0; i < xbin_nb; i++)
            hist.push_back((histT) 0.f);
        bounds[0] = x_bin_min;
        bounds[1] = x_bin_max;
    }

    /**
     * @brief setBinsNumber set the number of bins to x_nb
     * @param x_nb
     */
    void setBinsNumber( unsigned int x_nb )
    {
        xbin_nb = x_nb;
        hist.resize(xbin_nb);
    }

    /**
     * @brief setBounds set the minimum and maximum bounds
     * @param x_bin_min
     * @param x_bin_max
     */
    void setBounds(T  x_bin_min, T x_bin_max )
    {
        bounds[0] = x_bin_min;
        bounds[1] = x_bin_max;
    }

    /**
     * @brief addValue add the value in the bin corresponding to x_val
     * @param x_val to find the bin
     * @param value value to add in the bin
     */
    void addValue ( T x_val, histT value)
    {
        unsigned int x = (x_val-bounds[0]) / (bounds[1]-bounds[0]) * (T) (xbin_nb-1);
        assert(x < xbin_nb);
        hist[x] += value;
        values_sum += value;
    }

    /**
     * @brief clearValues
     */
    void clearValues(){
        for (unsigned int i = 0; i < xbin_nb; i++)
            hist[i]= 0;
    }

    /**
     * @brief distanceL2 computes the L2 distance between this and h2
     * @param h2
     * @return L2 distance
     */
    histT distanceL2( const Histogram1D & h2 ){
        histT res = 0.f;
        for (unsigned int i = 0; i < xbin_nb; i++)
            res += (hist[i]/values_sum-h2.hist[i]/h2.values_sum)*
                    (hist[i]/values_sum-h2.hist[i]/h2.values_sum);
        return sqrt(res);

    }

    /**
     * @brief distanceL1 computes the L1 distance between this and h2
     * @param h2
     * @return L1 distance
     */
    histT distanceL1( const Histogram1D & h2 ){
        histT res = 0.f;
        for (unsigned int i = 0; i < xbin_nb; i++)
            res += abs(hist[i]/values_sum-h2.hist[i]/h2.values_sum);
        return sqrt(res);
    }

    /**
     * @brief distanceChi2 compute the chi2 distance between two 1D histograms
     * @param h2
     * @return
     */
    histT distanceChi2(const Histogram1D & h2)
    {
        histT res = 0.f;
        assert( xbin_nb == h2.xbin_nb);
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            if ( hist[i]+ h2.hist[i] != 0 )
                res += (hist[i]-h2.hist[i])* (hist[i]-h2.hist[i])
                        /  (hist[i]+h2.hist[i]);
        }
        return res;
    }

    /**
     * @brief distanceChi2Normalized
     * @param h2
     * @return
     */
    histT distanceChi2Normalized(const Histogram1D & h2)
    {
        histT res = 0.f;
        assert( xbin_nb == h2.xbin_nb);
        histT valH1, valH2;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            valH1 = hist[i]/values_sum;
            valH2 = h2.hist[i]/h2.values_sum;
            if ( valH1+ valH2 != 0 )
                res += (valH1-valH2)* (valH1-valH2) / (valH1+valH2);
        }
        return res;
    }


    /**
     * @brief drawHistogram
     * @return to display the histogram in a qt window
     */
    QPixmap drawHistogram(){
        // Create an histogram_ and display it
        const int height = 150;
        const int top_margin = 5;
        const int left_margin = 15;
        const int drawing_height = height-top_margin*2;
        const int width = 202;
        const int cell_width = (width - 2*left_margin)/xbin_nb;
        const int text_margin = 3;
        const int text_height = 20;

        drawHist = QPixmap(width,height+text_height);
        drawHist.fill(QColor(192,192,192));

        QPainter painter(&drawHist);
        painter.setPen(Qt::black);
        painter.setBrush(QColor(128,128,128));
        //painter.setFont(QFont("Arial", 30));

        // Get maximum value (to normalize)
        double interval = fabs(bounds[1] - bounds[0]);
        double step = interval/xbin_nb;

        float max_size = 0.f;
        int m_id = 0;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            max_size = std::max(max_size, hist[i]);
            if (max_size == hist[i])
                m_id = i;
        }

        // colored histogram
        int j = 0;

        // draw
        int c=left_margin;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            int line_height ;
#if WINDOWS_USED
            line_height = static_cast<int>( round_W(static_cast<double>(drawing_height) *
                                                    static_cast<double>(hist[i])/static_cast<double>(max_size)));
#else
           // line_height = static_cast<int>( round(static_cast<double>(drawing_height) *
             //                                     static_cast<double>(hist[i])/static_cast<double>(max_size)));
#endif

            painter.fillRect(c,
                             drawing_height+top_margin-line_height,
                             cell_width,
                             line_height,
                             get_histogram_color(j++));
            c += cell_width;
        }

        // draw bottom horizontal line
        painter.setPen(Qt::blue);

        painter.drawLine(QPoint(left_margin, drawing_height + top_margin),
                         QPoint(left_margin + static_cast<int>(xbin_nb)*cell_width,
                                drawing_height + top_margin));


        // draw min value and max value

        const double max_nb_value = bounds[0] + m_id*step;


        const int tr_width = 30;
        const int min_tr_pos = static_cast<int>( left_margin );
        const int max_tr_pos = static_cast<int>( xbin_nb*cell_width + left_margin );
        const int max_nb_pos = static_cast<int>(
                m_id*cell_width + left_margin );
        const int tr_y = drawing_height + top_margin + text_margin;

        painter.setPen(get_histogram_color(bounds[0]));
        QRect min_text_rect (min_tr_pos - tr_width/2., tr_y, tr_width, text_height);
        QString text;
        text.setNum(bounds[0],'f',2);
        painter.drawText(min_text_rect, Qt::AlignCenter, text);

        painter.setPen(get_histogram_color(bounds[1]));
        QRect max_text_rect (max_tr_pos - tr_width/2., tr_y, tr_width, text_height);
        text.setNum(bounds[1],'f',2);
        painter.drawText(max_text_rect, Qt::AlignCenter, text);

        painter.setPen(get_histogram_color(max_nb_value));
        QRect max_nb_rect (max_nb_pos- tr_width/2., tr_y, tr_width, text_height);
        text.setNum(max_nb_value,'f',2);
        painter.drawText(max_nb_rect, Qt::AlignCenter, text);

        return drawHist;
    }

    /**
     * @brief get_histogram_color
     * @param v
     * @return
     */
    QColor get_histogram_color(const double v) const
    {
      if ( v < 5 )            { return QColor(215,108,0); }
      else if ( v < 10 )      { return QColor(215,108,0); }
      else if ( v < 15 )      { return QColor(138,139,0); }
      else if ( v < 165 )     { return Qt::darkGreen; }
      else if ( v < 170 )     { return QColor(138,139,1); }
      else if ( v < 175 )     { return QColor(215,108,0); }
      else                    { return Qt::red; }
    }


protected:
    std::vector< histT > hist; /**< histogram */
    unsigned int xbin_nb; /**< number of bins */
    T bounds[2]; /**< minimum and maximum bounds of the histogram*/
    QPixmap drawHist; /**< to display the histogram*/
    histT values_sum; /**< sum of all the values in the histogram*/

};
typedef Histogram1D<float, float> Hist1Df;


#endif // HISTOGRAM1D_H
