/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MATHTOOLS_H
#define MATHTOOLS_H

#ifdef _MSC_VER
#  define random rand
#  define NOMINMAX
#  ifdef min
#    undef min
#  endif
#  ifdef max
#    undef max
#  endif
#endif

#include <cstdlib>
#include <vector>
#include <algorithm>

#include <Vec3D.h>
#include <Vertex.h>

#define FLOAT_PRECISION 0.0000001

/**
 * @brief principalComponentAnalysis compute a PCA on the points
 * @param points (in)
 * @param eigenValues (out) eigen values
 * @param eigenVectors (out) principal direction vector
 */
void principalComponentAnalysis (const std::vector<Vec3Df> & points,
                                 float * eigenValues, Vec3Df * eigenVectors);

/**
 * @brief principalComponentAnalysisOnNormals compute a PCA on the normal of the points
 * @param vertices (in)
 * @param indices (in) take in account the vertice which index is in the indices vector
 * @param eigenValues (out) eigen values
 * @param eigenVectors (out) principal direction vector
 */
void principalComponentAnalysisOnNormals (const std::vector<Vertex> & vertices,
                                        const std::vector<unsigned int> & indices,
                                        float * eigenValues, Vec3Df * eigenVectors);

/**
 * @brief WeightedPCA compute a weighted PCA
 * @param datas (in)
 * @param weights (in) weights of the datas
 * @param eigenValues (out) eigen values
 * @param eigenVectors (out) principal direction vector
 * @return
 */
Vec3Df WeightedPCA(const std::vector<Vec3Df> & datas,
                  const std::vector<double> & weights,
                  float* eigenValues, Vec3Df* eigenVectors);


template<class T>
/**
 * @brief furthestPoint find the
 * @param p (in) current center
 * @param points_ind (in) points indices
 * @param ind (in) index to look at
 * @param points (in) datas
 * @param maxP (out) position of the furthest point
 * @param maxD (out) distance between the center and the furthest point
 */
void furthestPoint (const Vec3Df & p, const std::vector<T> & points_ind, T ind,
                    const std::vector<Vertex> & points, Vec3Df & maxP, float & maxD){
    // return if the point cloud is empty
    if (points.size () == 0) {
        maxP = p;
        maxD = 0;
        return;
    }

    bool found = false;
    unsigned int i = 0;
    while (!(found || i > points.size()))
    {
        if (points_ind[i] == ind)
        {
            maxP = points[i].getPos();
            found = true;
        }
        i++;
    }

    // return if there is no point to consider
    // ie no i such as points_ind[i] == ind
    if (!found)
        return;

    // search the maximum distance and the position of the associated point
    // between the points and the current center

    maxD = Vec3Df::squaredDistance (p, maxP);
    for (unsigned int i = 1; i < points.size (); i++) {
        if (points_ind[i] == ind)
        {
            float d = Vec3Df::squaredDistance (points[i].getPos(), p);
            if (d > maxD) {
                maxD = d;
                maxP = points[i].getPos();
            }
        }
    }
    maxD = sqrt (maxD);
}

template<class T>
/**
 * @brief miniball compute the miniball algorithm find the smallest sphere that encloses the datas
 * only the points with a points_ind == ind are considered
 * @param points_ind  (in) points indices
 * @param ind index to look at
 * @param points (in) points
 * @param center (out) center of the miniball
 * @param maxD (out) radius of the miniball
 * @param epsilonPrecision precision used to stop the algorithm
 */
void miniball (const std::vector<T> & points_ind, T ind,
               const std::vector<Vertex> & points, Vec3Df & center, float & maxD, double epsilonPrecision = 0.1){
    // if the point cloud has only 1 point or less
    // the center is the point, the radius is 0
    if (points.size () <= 1) {
        if (points.size () == 1)
        {
            if (points_ind[0] == ind)
                center = points[0].getPos();
        }
        maxD = 0;
        return;
    }

    Vec3Df c;
    bool found = false;
    unsigned int i = 0;
    while (!(found || i > points.size()))
    {
        if (points_ind[i] == ind)
        {
            c = points[i].getPos();
            found = true;
        }
        i++;
    }

    // we return if there is no point to consider
    // ie no i such as points_ind[i] == ind
    if (!found)
    {
        maxD = 0;
        return;
    }

    unsigned int limit = static_cast<unsigned int>(1.0/(epsilonPrecision*epsilonPrecision));
    // find the furthest point and update the center
    for (unsigned int i = 2; i < limit; i++) {
        Vec3Df q;
        furthestPoint (c, points_ind, ind, points, q, maxD);
        c = c + (q - c) * (1.0/i);
    }
    center = c;
}

#endif // MATHTOOLS_H

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:

