/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "GLViewer.h"

#include <QCoreApplication>
#include <QMouseEvent>

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <string>
#include <algorithm>

#define USE_TEXTURE_TO_DEBUG false
#include "GLUtilityMethods.h"

#include "CotanLaplacian.h"

using namespace std;

//***************************************************
//               Mesh I/O
//***************************************************

void GLViewer::open (const QString & filename) {
    _widget->setRadius(0.02f);
    if (filename.endsWith(".off"))
        openOFF (string (qPrintable (filename)), 2, mesh);
    else if (filename.endsWith(".obj")){
        openOBJ (string (qPrintable (filename)), 2, mesh);
    } else {
        Trace(1," issue, unknown mesh format")
        return;
    }
    mesh.computeMeshProperties();
    _selectionsManager->init(mesh);
    meshInitialized = true;
    emit resetStrokeToDraw();
    updateGL();
}

void  GLViewer::saveMesh (const QString & filename)
{
    const vector<Vertex> & V = mesh.getVertices ();
    const vector<Triangle> & T =  mesh.getTriangles ();
    const std::string file_n = qPrintable(filename);
    ofstream out (file_n.c_str ());

    if (!out)
       exit (EXIT_FAILURE);

    out << "OFF \n";
    out << V.size() << " " << T.size() << " 0\n";

    Vec3Df pos;
    for (unsigned int i= 0; i < V.size(); i++)
    {
        pos = V[i].getPos();
        out << pos << "\n";
    }

    for (unsigned int i = 0; i < T.size(); i++) {
        out << "3 ";
        for (unsigned int j = 0; j < 3; j++)
            out << T[i].getVertex(j) << " ";
        out << "\n";
    }
    out.close();
}

//***************************************************
//                  selections
//***************************************************

void GLViewer::saveSelections(const QString &filename){
    ofstream out (filename.toUtf8());
    if (!out)
        exit (EXIT_FAILURE);

    Trace (1, " saveSelections ");
    const vector<float> & selections = _selectionsManager->getSelections();

    out << selections.size();
    out << endl;

    for (unsigned int i = 0; i < selections.size(); i++)
        out << selections[i] << endl;
    out.close ();
}

void GLViewer::loadSelections(const QString &filename){
    std::ifstream file;
#if (QT_VERSION < 5)
    file.open(filename.toAscii());
#else
    file.open(filename.toLatin1());
#endif
    vector<float> selections;

    unsigned int nb_triangles;
    file >> nb_triangles;
    if (nb_triangles != mesh.getTriangles().size())
    {
        Trace(1, " ISSUE, you are not loading a corresponding selections file. ");
        for (unsigned int i = 0; i < mesh.getTriangles().size(); i++)
            selections.push_back(0.f);
        _selectionsManager->setSelections(selections);
        updateGL();
        return;
    }
    float sel;
    for (unsigned int i = 0; i< nb_triangles; i++)
    {
        file >> sel;
        selections.push_back(sel);
    }
    _selectionsManager->setSelections(selections);
    updateGL();
}

void GLViewer::setSelectionType(TYPE t){
    _widget->clear();
    _selectionsManager->clearCurrentSelections();
    _selectionsManager->setSelectionType(t);
    emit selectionTypeChange(t);
}

void GLViewer::setEraserMode(bool b){
    Trace(2, "GLViewer::setEraseWidgetMode");
    if (b)
        _widget->setMode(1);
    else
        _widget->setMode(0);
}

void GLViewer::setDeleteSelectionMode(bool b){
    Trace(2, "GLViewer::setDeleteStrokeMode");
    if (b)
        _widget->setMode(2);
    else
        _widget->setMode(0);
}

void GLViewer::setDrawStrokeMode(bool b){
    if (b){
        _widget->setMode(3);
    }
    else
        _widget->setMode(0);
}

void GLViewer::saveTimings(const QString &filename){
    _selectionsManager->saveTimings(filename);
}

//***************************************************
//                  expansion
//***************************************************

void GLViewer::openSVMselection(const QString & filename)
{
    _selectionsManager->openSVMselection(filename);
    updateGL();
}

QPixmap GLViewer::getSCRefHistogram(){
   return _selectionsManager->getRefHistogram();
}

QPixmap GLViewer::getSCCurHistogram(float & dist){
    return _selectionsManager->getUnderMouseHistogram(vertexUnderMouse, dist);
}

//***************************************************
//                  cameras & video
//***************************************************

void GLViewer::setVideoName(const QString &filename) {
    setSnapshotFileName(filename);
    Trace(1, " FILE NAME FOR SNAP SHOT "<< qPrintable (filename));
}

void GLViewer::saveCamera(const QString &filename){
    ofstream out (filename.toUtf8());
    if (!out)
        exit (EXIT_FAILURE);

	out << camera()->position().x << " " <<
			camera()->position().y <<" " << 
			camera()->position().z <<" " <<
            camera()->viewDirection().x << " " <<
            camera()->viewDirection().y << " " <<
            camera()->viewDirection().z << " " <<
            camera()->upVector().x << " " <<
            camera()->upVector().y << " " <<
            camera()->upVector().z << " " <<
            camera()->fieldOfView();
    out << endl;
    out.close ();
}

std::istream & operator>>(std::istream & stream, qglviewer::Vec & v)
{
    stream >>
            v.x >>
            v.y >>
            v.z;

    return stream;
}

void GLViewer::openCamera(const QString &filename){

    std::ifstream file;
#if (QT_VERSION < 5)
    file.open(filename.toAscii());
#else
    file.open(filename.toLatin1());
#endif
    qglviewer::Vec pos;
    qglviewer::Vec view;
    qglviewer::Vec up;
    float fov;

    file >> pos >>
            view >>
            up >>
            fov;

    camera()->setPosition(pos);
    camera()->setViewDirection(view);
    camera()->setUpVector(up);
    camera()->setFieldOfView(fov);

    camera()->computeModelViewMatrix();
    camera()->computeProjectionMatrix();

    updateGL();
}



//***************************************************
//               Viewer protected functions
//***************************************************

void GLViewer::init () {
    Trace(2, "GLViewer::init");
    _beginvideo = false;
    debug = false;
    _debugVisuMode = NO_VISUMODE;
    debugTextureId = 0;

  glewInit();
  if (glewGetExtension ("GL_ARB_vertex_shader")        != GL_TRUE ||
      glewGetExtension ("GL_ARB_shader_objects")       != GL_TRUE ||
      glewGetExtension ("GL_ARB_shading_language_100") != GL_TRUE) {
      cerr << "Driver does not support OpenGL Shading Language" << endl;
      exit (EXIT_FAILURE);
  }
  if (glewGetExtension ("GL_ARB_vertex_buffer_object") != GL_TRUE) {
      cerr << "Driver does not support Vertex Buffer Objects";
      exit (EXIT_FAILURE);
  }
  setBackgroundColor (Qt::white);
  setAutoBufferSwap (true);
  glCullFace (GL_BACK);
  glEnable (GL_CULL_FACE);
  glDepthFunc (GL_LESS);
  glEnable (GL_DEPTH_TEST);

  initTexture();
  initLights ();
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT2);
  glEnable(GL_LIGHT4);
  setDefaultMaterial ();

  wireframe = false;

  _strokeColor = QColor( 0.f, 50.f, 235.f);
  _patchColor = QColor( 222.f, 156.f, 0.f);
  _partColor = QColor( 54.f, 140.f, 23.f);
  _ccColor = QColor(7.f, 113.f, 162.f);
  _bgColor = QColor(83.f, 83.f, 83.f);
  _csColor = QColor(183.f, 0.f, 0.f);
  _negColor = QColor(248.f, 94.f, 0.f);

  // shader
  try {
      phongShader = new PhongShader ( "./shaders/simSelect_shader.vert", "./shaders/simSelect_shader.frag" );
      phongShader->loadFromFile("./shaders/simSelect_shader.vert", "./shaders/simSelect_shader.frag");
      phongShader->bind ();
      setShaderValues ();
      phongShader->unbind ();
  } catch (ShaderException e) {
      std::cout << " shader exception catched \n";
      cerr << e.getMessage () << endl;
      exit (EXIT_FAILURE);
  }

  // connection to signal widget
  connect(_widget, SIGNAL(pressed()), this, SLOT(startSelection()) );
  connect(_widget, SIGNAL(moved()), this, SLOT(updateSelection()));
  connect(_widget, SIGNAL(released()), this, SLOT(endSelection()));
  connect(_widget, SIGNAL(released()), this, SLOT(selectionTypeImgNeedUpdate()));

  setMouseTracking(true);
  setSnapshotQuality(100);

  // Add custom key description (see keyPressEvent).
  setKeyDescription(Qt::Key_W, "Toggles wireframe display");
  setKeyDescription(Qt::Key_E, "Toggles the selection mode");
  setKeyDescription(Qt::Key_Return, "Accept current selections");
  setKeyDescription(Qt::Key_P, "Toggles patch selection mode");
  setKeyDescription(Qt::Key_M, "Toggles part selection mode");
  setKeyDescription(Qt::Key_C, "Toggles components selection mode");
  setKeyDescription(Qt::Key_S, "Find Similar selections");
  setKeyDescription(Qt::Key_Tab, "Toggles debug mode");

  setMouseBindingDescription(Qt::NoModifier, Qt::LeftButton,
                             " on the model : performs selection (if selection ON) <br/> on the background : rotates camera");


  setMouseBindingDescription(Qt::NoModifier, Qt::RightButton,
                             " on the model : erases selection (if selection ON) <br/> on the background : translates camera");

  setMouseBindingDescription(Qt::NoModifier, Qt::MiddleButton,
                             " on the model : changes the tool size (if sel. ON) <br/> on the background : zooms camera");
}

void GLViewer::draw ()
{
    initLights();

    glEnable (GL_COLOR_MATERIAL);
    setDefaultMaterial();

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    if (wireframe)
        glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
    else
        glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

    if (debug)
        drawMeshDebug();
    else
    {
        phongShader->bind();
        setShaderValues();
        glColor3f (0.7, 0.7, 0.7);
        drawMesh();
        phongShader->unbind();
    }

    vector<float> color;
    color.push_back(_strokeColor.redF());
    color.push_back(_strokeColor.greenF());
    color.push_back(_strokeColor.blueF());

    if (debug)
    {
        switch (_debugVisuMode){
        case SIMILARITY_FIELD:
            _selectionsManager->drawPartSimilarityConstraints();
            break;
        case USER_FIELD :
            if (_selectionsManager->getCurrentSelectionType() == PATCH)
                _selectionsManager->drawPatchSCDebug();

            _selectionsManager->drawPartIsolines();
            _selectionsManager->drawPartSimilarityConstraints();
            _selectionsManager->drawPartUserConstraints();
            break;
        case NO_VISUMODE:
            _selectionsManager->drawPartIsolines();
            _selectionsManager->drawPartSimilarityConstraints();
            _selectionsManager->drawPartUserConstraints();
            if (_selectionsManager->getCurrentSelectionType() == PATCH)
                _selectionsManager->drawPatchSCDebug();
            if (_selectionsManager->getCurrentSelectionType() == PATCH)
                _selectionsManager->drawPatchICPDebug();
        default:
            break;
        }
    }

    if (_selectionsManager->getCurrentSelectionType() == PART)
        _widget->draw(color, true);
    else
        _widget->draw(color, false);

    if (_beginvideo)
        saveSnapshot();
}

QString GLViewer::helpString() const {

  QString text("<h2> SimSelect: Similarity-based Selection for 3D Surfaces  </h2>");
  text += "<h3> Version 0.1 </h3> ";

  text += "<hr>";
  text += "<h3> Authors:</h3> Emilie Guy, Jean-Marc Thiery, Tamy Boubekeur";
  text += "<p>(http://perso.telecom-paristech.fr/~guy)</p>";
  text += "<hr>";
  text += "<p>SimSelect is an interactive graphics application allowing to perform advanced selections on 3D meshes.";
  text += "Three types of selections are available: components, parts and ";
  text += "patches. When defining them, SimSelect automatically adjusts their ";
  text += "boundaries while the user only performs rough gestures. ";
  text += "Once a selection is defined, similar selections can be found ";
  text += "automatically on the entire model. The selection set can manually be";
  text += "adjusted and saved in a simple text format. </p> <br/>";

  text += " For more help on the usage, please have a look at the \"Keyboard\" and \"Mouse\" tabs ";
  text += " Other tooltips are displayed when you hover the buttons. ";

//  text += "<p> SimSelect is a tool to assist you during the selection process. </p>";
//  text+=  "<p> We designed 3 types of selections: connected components, parts or patches.  ";
//  text += " Click to select connected components, cut along the boundaries of parts and brush over patches (change brush radius using the mouse wheel). </p>";
//  text += "<p> To further assist you, we add an <b> expansion </b> tool <br/>";
//  text += " Use it to detect similar selections: click on the expansion button (right top) and change the number of similar selections using the slider. </p> ";

  text += "<br/>";
  text += "<h3> In / Out: </h3> ";
  text += "<hr>";
  text += "<p>Input : meshes ( OFF (.off), and OBJ (.obj) formats are currently supported)</p>";
  text += "<p>Output : selections (.seg files)</p>";
  text +=  "A .seg file is structured as follow:<br/>";
  text += " <br/>";
  text += "number_of_faces <br/>";
  text += "list of selection indices (one for each face of the input model) <br/>";
  text += "  <br/>";
  text += "<br/>";
  text += "the selection indices are float and have the following meaning:";
  text += "<ul>";
  text +="<li> &lt; 0 : no selection</li>";
  text += "<li>0 &lt; 10 : patches</li>";
  text += "<li>10 &lt;  100 : parts</li>";
  text += "<li>100 &lt; 1000 : components</li>";
  text += " </ul>";
  text += "<p>the reference selection is an int. The similar selections are float where ";
  text += "the int part is the same as the reference selection and the decimal one is";
  text += "the distance between the local selection and the reference one </p>";
  text += "</ul>";
  text += "<br/>";
  text += " <h2> CITATION </h2>";
  text += "<hr>";
  text += " <p>Please cite the following paper when using SimSelect: </p>";

  text += " <p> <i> SimSelect: Similarity-based Selection for 3D Surfaces  ";
  text += " Emilie Guy, Jean-Marc Thiery and Tamy Boubekeur";
  text += " Computer Graphics Forum (Proc. EUROGRAPHICS 2014), volume 33, issue 2, pp.165-173, 2014. </i> </p>";

  text += "<br/>";

  text += " <h2> Credits and License </h2>";
  text += "<hr>";
  text += " SimSelect - Similarity-based Selection for 3D Surfaces  <br/>";
  text += "Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR<br/>";
  text += "Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom<br/>";
  text += "    All rights reserved<br/>";
  text += "<p>SimSelect is free software: you can redistribute it and/or modify ";
  text += "it under the terms of the GNU Lesser General Public License as published by ";
  text += "the Free Software Foundation, either version 3 of the License, or ";
  text += "(at your option) any later version. ";
  text += "SimSelect is distributed in the hope that it will be useful, ";
  text += "but WITHOUT ANY WARRANTY; without even the implied warranty of ";
  text += "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the ";
  text += "GNU Lesser General Public License for more details. ";
  text += "You should have received a copy of the GNU Lesser General Public License";
  text += "along with this program. If not, see <http://www.gnu.org/licenses/>. </p>";


  return text;
}

//***************************************************
//                  events
//***************************************************

bool GLViewer::event(QEvent *event)
{
    if (event->type() == QEvent::KeyPress) {
        QKeyEvent *ke = static_cast<QKeyEvent *>(event);
        if (ke->key() == Qt::Key_Tab) {
            debug = !debug;
            emit ( debugMode(debug));
            updateGL();
            return true;
        }
    }
    return QWidget::event(event);
}

void GLViewer::keyPressEvent (QKeyEvent * e) {

    if (e->text() == "w")
        wireframe = ! wireframe;
    else if (e->text() == "z")
    {
        _beginvideo = !_beginvideo;
        if ( _beginvideo)
            emit signalTextToStatusBar(" you are recording a video.. press z to finish it");
        else
            emit signalTextToStatusBar(" ");
    }
    QGLViewer::keyPressEvent (e);
}


bool GLViewer::clicToPoint3D(int currentX, int currentY, Vec3Df & p)
{
    float z;
   glReadBuffer (GL_FRONT);
   glReadPixels (currentX, height () -currentY-1, 1, 1,
                 GL_DEPTH_COMPONENT, GL_FLOAT, &z);
   qglviewer::Vec proj =
           camera ()->unprojectedCoordinatesOf (qglviewer::Vec (currentX, currentY, z));
   p = Vec3Df (proj.x, proj.y, proj.z);
   return (z != 1);
}

void GLViewer::mousePressEvent(QMouseEvent* e) {
    QGLViewer::mousePressEvent(e);
}

void GLViewer::mouseMoveEvent(QMouseEvent *e){
    if (meshInitialized)
    {
        bool found;
        qglviewer::Vec point = camera()->pointUnderPixel(e->pos(), found);

        if (found && ( (_selectionsManager->getCurrentSelectionType()==PATCH) ||
                       (_selectionsManager->getCurrentSelectionType()==PART) ||
                       (_selectionsManager->getCurrentSelectionType()==COMP)) && debug)
        {
            int v = mesh.vertexCloseToPoint(Vec3Df(point.x, point.y, point.z), 0.1);
            if ( v != -1)
            {
                vertexUnderMouse = v;
                emit updateScCurHistogram();
            }
        }
    }
    _widget->setRepUplan(Vec3Df(camera()->rightVector()));
    _widget->setRepVplan(Vec3Df(camera()->upVector()));
    QGLViewer::mouseMoveEvent(e);
}

void GLViewer::mouseReleaseEvent(QMouseEvent *e){
    QGLViewer::mouseReleaseEvent(e);
}

void GLViewer::wheelEvent(QWheelEvent *e){
    QGLViewer::wheelEvent(e);
}

//***************************************************
//                  mesh I/O
//***************************************************

void GLViewer::openOFF (const std::string filename, unsigned int normWeight, Mesh & mesh){
    mesh.clear ();
    std::vector<Vec3Df>  positions ;
    vector<Vertex> & _V = mesh.getVertices ();
    vector<Triangle> & _T = mesh.getTriangles ();
    ifstream in (filename.c_str ());
    FileIO::openOFF(in, positions, _T);
    for (unsigned int i = 0; i < positions.size(); i++) {
        _V.push_back (Vertex (positions[i]));
    }
    for (unsigned int i = 0; i < _T.size(); i++)
    {
        _V[_T[i].getVertex(0)].addFace(i);
        _V[_T[i].getVertex(1)].addFace(i);
        _V[_T[i].getVertex(2)].addFace(i);
    }
    Vec3Df center;
    float radius;
    Vertex::scaleToUnitBox (_V, center, radius);
    mesh.recomputeSmoothVertexNormals (normWeight);
}

void GLViewer::openOBJ(const std::string filename, unsigned int normWeight, Mesh & mesh){
    mesh.clear ();
    std::vector<Vec3Df>  positions ;
    vector<Vertex> & _V = mesh.getVertices ();
    vector<Triangle> & _T = mesh.getTriangles ();
    ifstream in (filename.c_str ());
    FileIO::openOBJ(in, positions, _T);
    for (unsigned int i = 0; i < positions.size(); i++) {
        _V.push_back (Vertex (positions[i]));
    }
    for (unsigned int i = 0; i < _T.size(); i++)
    {
        _V[_T[i].getVertex(0)].addFace(i);
        _V[_T[i].getVertex(1)].addFace(i);
        _V[_T[i].getVertex(2)].addFace(i);
    }
    Vec3Df center;
    float radius;
    Vertex::scaleToUnitBox (_V, center, radius);
    mesh.recomputeSmoothVertexNormals (normWeight);
}

void GLViewer::initTexture()
{
    glEnable(GL_TEXTURE_2D);

    glGenTextures(1, ui_text_id );

    QImage textimg , buf;
    buf.load("./GUI/texture_sim.png");
    textimg = QGLWidget::convertToGLFormat( buf );

    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_MAG_FILTER , GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_MIN_FILTER , GL_LINEAR );

    ui_text_id[0] = this->bindTexture( textimg, GL_TEXTURE_2D, GL_RGB );

    buf.load("./GUI/texture.png");
    textimg = QGLWidget::convertToGLFormat( buf );

    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_MAG_FILTER , GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D , GL_TEXTURE_MIN_FILTER , GL_LINEAR );

    ui_text_id[1] = this->bindTexture( textimg, GL_TEXTURE_2D, GL_RGB );

    glDisable(GL_TEXTURE_2D);
}

//***************************************************
//                    Lights & Drawing Fns
//***************************************************
void GLViewer::initLights () {

    GLfloat light_position0[4] = {-50, 50, -10, 0};
    GLfloat light_position2[4] = {473, -351, -259, 0};
    GLfloat light_position4[4] =  { 10, 10, 50, 0};

    GLfloat direction2[3] = {-473, 351, 259};
    GLfloat direction4[3] = {-10, -10, -50};

    GLfloat color2[4] = {0.28, 0.39, 0.50, 1};
    GLfloat color4[4] = {0.3, 0.2, 0.1, 1};

    GLfloat specularColor2[4] = {0.8, 0.8, 0.8, 1};
    GLfloat specularColor3[4] = {0.8, 0.8, 0.8, 1};
    GLfloat ambient[4] = {0.3f, 0.3f, 0.3f, 0.5f};

    glLightfv (GL_LIGHT0, GL_POSITION, light_position0);

    glLightfv (GL_LIGHT2, GL_POSITION, light_position2);
    glLightfv (GL_LIGHT2, GL_SPOT_DIRECTION, direction2);
    glLightfv (GL_LIGHT2, GL_DIFFUSE, color2);
    glLightfv (GL_LIGHT2, GL_SPECULAR, specularColor2);

    glLightfv (GL_LIGHT4, GL_POSITION, light_position4);
    glLightfv (GL_LIGHT4, GL_SPOT_DIRECTION, direction4);
    glLightfv (GL_LIGHT4, GL_DIFFUSE, color4);
    glLightfv (GL_LIGHT4, GL_SPECULAR, specularColor3);
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambient);

    glEnable (GL_LIGHTING);
}


void GLViewer::setDefaultMaterial () {

    GLfloat material_color[4] = {1, 1, 1, 1.0};
    GLfloat material_specular[4] = {0.5,0.5,0.5,1.0};
    GLfloat material_ambient[4] = {0.0,0.0,0.0,1.0};

    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_color);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_ambient);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128);

    glEnable(GL_COLOR_MATERIAL);
    diffuseRef = 0.7f; // 0.8
    specRef =  0.3f; //1.5
    shininess = 1.f;// 128.0f;
}

void GLViewer::setShaderValues ( ) {
    phongShader->setDiffuseRef (diffuseRef);
    phongShader->setSpecRef (specRef);
    phongShader->setShininess (shininess);
}

void GLViewer::drawMesh () {
    const vector<Vertex> & V = mesh.getVertices ();
    const vector<Triangle> & T = mesh.getTriangles ();

    const vector<float> & currentSelections =  _selectionsManager->getCurrentSelections();
    const vector<float> & selections = _selectionsManager->getSelections();

    glBegin (GL_TRIANGLES);

    for (unsigned int i = 0; i < T.size (); i++) {
        const Triangle & t = T[i];
        QColor c_base;
        if (currentSelections[i] != -1.f) {
            float dist = currentSelections[i]-(int) currentSelections[i] ;
            if (dist == 0.f)
                glColor4f(_csColor.redF(), _csColor.greenF(), _csColor.blueF(), 0.8f);
            else
            {
                QColor c_base = QColor(0.8*255, 0.3*255, 0.3*255);
                QColor c_hsv = c_base.toHsv();
                c_hsv.setHsvF( c_hsv.hueF(), c_hsv.saturationF()*(1-dist), c_hsv.valueF() );
                c_base = c_hsv.toRgb();
                glColor4f( c_base.redF(), c_base.greenF(), c_base.blueF(),0.8 );
            }
        } else if (selections[i] != -1.f)
        {
            if (selections[i] >= 1000.f)
                c_base = _partColor;
            else if (selections[i] >= 100.f)
                c_base = _ccColor;
            else if ( (selections[i] >= 1.f) || (selections[i] == -4.f ))
                c_base = _patchColor;

            float dist = selections[i]-(int) selections[i] ;
            if (dist == 0.f)
                glColor4f(c_base.redF(), c_base.greenF(), c_base.blueF(), 0.8f);
            else
            {
                QColor c_hsv = c_base.toHsv();
                c_hsv.setHsvF( c_hsv.hueF(), c_hsv.saturationF()*(1-dist), c_hsv.valueF() );
                c_base = c_hsv.toRgb();
                glColor4f( c_base.redF(), c_base.greenF(), c_base.blueF(),0.8 );
            }
        }
        else {
            glColor4f( _bgColor.redF(), _bgColor.greenF(), _bgColor.blueF(), 0.f );
        }
        for (unsigned int j = 0; j < 3; j++)
        {
            glNormalVec3Df (V[t.getVertex (j)].getNormal());
            glVertexVec3Df (V[t.getVertex (j)].getPos ());

        }
    }
    glEnd ();
}

void GLViewer::drawMeshDebug () {
    const vector<Vertex> & V = mesh.getVertices ();
    const vector<Triangle> & T = mesh.getTriangles ();

    TYPE t = _selectionsManager->getCurrentSelectionType();

    vector<float> visuField;
    switch (_debugVisuMode)
    {
    case USER_FIELD :
        glBindTexture( GL_TEXTURE_2D , ui_text_id[debugTextureId]);
        glEnable( GL_TEXTURE_2D );
        visuField = _selectionsManager->getUserField();
        break;
    case SIMILARITY_FIELD:
        glBindTexture( GL_TEXTURE_2D , ui_text_id[debugTextureId] );
        glEnable( GL_TEXTURE_2D );
        if ( t == PATCH)
            visuField = _selectionsManager->getPatchSimField();
        else if (t == COMP)
            visuField = _selectionsManager->getCompSimField();
        else if (t == PART)
            visuField = _selectionsManager->getConformalFactor();
        else
        {
            for (unsigned int i = 0; i < V.size(); i++)
                visuField.push_back(0.f);
        }
        break;
    case NO_VISUMODE:
        phongShader->bind();
        setShaderValues();
        glColor4f (0.7, 0.7, 0.7, 0.f);
        visuField = _selectionsManager->getCurrentSelections();
        break;
    default:
        phongShader->bind();
        setShaderValues();
        glColor4f (0.7, 0.7, 0.7, 0.f);
        break;
    }

    glBegin (GL_TRIANGLES);

    for (unsigned int i = 0; i < T.size (); i++) {
        const Triangle & t = T[i];
        for (unsigned int j = 0; j < 3; j++)
        {
            switch (_debugVisuMode){

            case USER_FIELD :
                glColor4f(1.f, 1.f, 1.f, 1.f);
                glTexCoord2f( visuField[t.getVertex (j)], 0.0 );
                break;
            case SIMILARITY_FIELD:
                glColor4f(1.f, 1.f, 1.f, 1.f);
                glTexCoord2f(1.f - visuField[t.getVertex (j)], 0.0 );
                break;
            case NO_VISUMODE:
                if (visuField[i] >= 0.f)
                {
                    float dist = visuField[i]-(int) visuField[i] ;
                    if (dist == 0.f)
                         glColor4f(_csColor.redF(), _csColor.greenF(), _csColor.blueF(), 0.8f);
                    else
                    {
                        QColor c_base = QColor(0.8*255, 0.3*255, 0.3*255);
                        QColor c_hsv = c_base.toHsv();
                        c_hsv.setHsvF( c_hsv.hueF(), c_hsv.saturationF()*(1.f-dist), c_hsv.valueF() );
                        c_base = c_hsv.toRgb();
                        glColor4f( c_base.redF(), c_base.greenF(), c_base.blueF(),0.8 );
                    }
                } else if (visuField[i] == -4.f)
                    glColor4f( _negColor.redF(), _negColor.greenF(), _negColor.blueF(), 1.f );
                else if (visuField[i] == -2.f)
                    glColor4f( _partColor.redF(), _partColor.greenF(), _partColor.blueF(), 1.f );
                else
                    glColor4f( _bgColor.redF(), _bgColor.greenF(), _bgColor.blueF(), 0.f );
                break;
            default:
                break;
            }
            glNormalVec3Df (V[t.getVertex (j)].getNormal());
            glVertexVec3Df (V[t.getVertex (j)].getPos ());
        }
    }
    glEnd ();

    switch (_debugVisuMode){
    case USER_FIELD :
        glDisable(GL_TEXTURE_2D);
        glEnable(GL_COLOR_MATERIAL);
        break;
    case SIMILARITY_FIELD:
        glDisable(GL_TEXTURE_2D);
        glEnable(GL_COLOR_MATERIAL);
        break;
    case NO_VISUMODE:
        phongShader->unbind();
        break;
    default:
        phongShader->unbind();
        break;
    }
}

//***************************************************
//                   selection SLOTS
//***************************************************
void GLViewer::startSelection(){
    _selectionsManager->startSelection(_widget);
    updateGL();
}

void GLViewer::updateSelection(){
    _selectionsManager->updateSelection(_widget);
    updateGL();
}

void GLViewer::endSelection(){
    _selectionsManager->endSelection(_widget);
    if ((_selectionsManager->getCurrentSelectionType() == PATCH ||
         _selectionsManager->getCurrentSelectionType() == COMP) && _debugVisuMode)
        emit updateScRefHistogram();
    updateGL();
}

void GLViewer::setSelection(bool b){
    Trace(2, "GLViewer::setSelection");
    if (!b)
    {
        _widget->clear();
        _selectionsManager->clearCurrentSelections();
    }
    _widget->setState(b);
    updateGL();
}

void GLViewer::acceptSelections(){
    Trace(2, "GLViewer::acceptSelections");
    _widget->clear();
    emit selectionTypeChange(_selectionsManager->getCurrentSelectionType());
    _selectionsManager->saveCurrentSelections();
    updateGL();
}

void GLViewer::selectionTypeImgNeedUpdate(){
    TYPE t = _selectionsManager->getCurrentSelectionType();
    emit selectionTypeChange(t);
}

//***************************************************
//                   visu SLOTS
//***************************************************

void GLViewer::updateStrokeColor(QColor c){
    _strokeColor = c;
    updateGL();
}

void GLViewer::updatePatchColor(QColor c){
    _patchColor = c;
    updateGL();
}

void GLViewer::updatePartColor(QColor c){
    _partColor = c;
    updateGL();
}

void GLViewer::updateccColor(QColor c){
    _ccColor = c;
    updateGL();
}

void GLViewer::updatebgColor(QColor c){
    _bgColor = c;
    updateGL();
}

void GLViewer::changeVisuMode(int c){
      _debugVisuMode = (VISU_MODE) c;
      updateGL();
}

void GLViewer::changeDebugTexture(int val){
    debugTextureId = val;
    updateGL();
}

//***************************************************
//                  Expansion  SLOTS
//***************************************************

void GLViewer::findSimilarSelections(){
    _selectionsManager->searchSimilarSelections(_widget->getStroke());
    if ((_selectionsManager->getCurrentSelectionType() == PATCH ||
         _selectionsManager->getCurrentSelectionType() == PART ||
         _selectionsManager->getCurrentSelectionType() == COMP) && _debugVisuMode)
        emit updateScRefHistogram();
    updateGL();
}


void GLViewer::changePatchSimAlgo(int i){
    _selectionsManager->changePatchSimAlgo( (PATCH_SIM_ALGO) i, _widget->getStroke());
    updateGL();
}

void GLViewer::thresholdMSChanged(int val){
    _selectionsManager->changeSimilarityThreshold(val);
    updateGL();
}
