/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "SelectionDock.h"

SelectionDock::SelectionDock(/*const QString & title,*/ GLViewer* viewer, QWidget* parent):QDockWidget(/*title,*/ parent)
{
    _viewer = viewer;
    this->setTitleBarWidget(0);
    QWidget * contents = new QWidget();
    QHBoxLayout * contentLayout = new QHBoxLayout(contents);
    QVBoxLayout* vBoxSel = new QVBoxLayout();
    contentLayout->addLayout(vBoxSel);

    QFont font;
    font.setPointSize(12);
    font.setCapitalization(QFont::SmallCaps);

    //***************************************************
    //        Actions on selections ToolBar
    //***************************************************

    QLabel* currentSelMan = new QLabel( "Selection ");
    currentSelMan->setFont(font);
    currentSelMan->setAlignment(Qt::AlignCenter);
    vBoxSel->addWidget(currentSelMan);

    QToolBar* setSelection = new QToolBar;
    select = new QAction( QIcon("./GUI/icons/brush.png") , "Set Selection", this);
    select->setShortcut(Qt::Key_E);
    select->setCheckable(true);
    select->setChecked(false);
    select->setToolTip("Set Selection (E)");
    connect(select, SIGNAL(toggled(bool)), _viewer, SLOT(setSelection(bool)));
    connect(select, SIGNAL(toggled(bool)), this, SLOT(activateStrokeMode(bool)));

    connect( _viewer, SIGNAL(resetStrokeToDraw()), this, SLOT(resetStrokeToDraw()));
    setSelection->addAction(select);

    addSelections = new QAction( QIcon("./GUI/icons/ok2.png") , "Accept current selections", this);
    addSelections->setShortcut(Qt::Key_Return);
    addSelections->setCheckable(false);
    addSelections->setStatusTip("Accept current selections (Enter)");
    connect(addSelections, SIGNAL(triggered()), _viewer, SLOT(acceptSelections()));
    setSelection->addAction(addSelections);

    reverseSelection = new QAction( QIcon("./GUI/icons/reverse2.png") , "reverse selection", this);
    reverseSelection->setCheckable(false);
    reverseSelection->setChecked(false);
    reverseSelection->setStatusTip("reverse selection");
    connect(reverseSelection, SIGNAL(triggered()), _viewer, SLOT(reverseCurrentSelection()));
    setSelection->addAction(reverseSelection);

    setSelection->setIconSize(QSize(32,32));
    vBoxSel->addWidget(setSelection);

    //***************************************************
    //            Selection Types ToolBar
    //***************************************************

    QFrame* line = new QFrame();
    line->setFrameShape(QFrame::VLine);
    line->setFrameShadow(QFrame::Sunken);
    contentLayout->addWidget(line, 1);

    QVBoxLayout* vBoxTypeSel = new QVBoxLayout();
    contentLayout->addLayout(vBoxTypeSel);

    QLabel* selType = new QLabel("Type");
    selType->setFont(font);
    selType->setAlignment(Qt::AlignCenter);
    vBoxTypeSel->addWidget(selType,1);
    QToolBar* chooseSelectionType = new QToolBar;

    QActionGroup* selectionTypeGroup = new QActionGroup(this);
    selectionTypeGroup->setExclusive(true);
    connect(selectionTypeGroup, SIGNAL(triggered(QAction*)),this, SLOT(setSelectionType(QAction*)));

    QAction * patch = new QAction( QIcon("./GUI/icons/patch2.png") , "Patch Selection", this);
    patch->setShortcut(Qt::Key_P);
    patch->setCheckable(true);
    patch->setChecked(true);
    patch->setToolTip("Patch Selection (p)");
    selectionTypeGroup->addAction(patch);

    QAction * part = new QAction( QIcon("./GUI/icons/part2.png") , "Part Selection", this);
    part->setShortcut(Qt::Key_M);
    part->setCheckable(true);
    part->setChecked(false);
    part->setToolTip("Part Selection (m)");
    selectionTypeGroup->addAction(part);

    QAction * comp = new QAction( QIcon("./GUI/icons/comp2.png") , "Connected Component Selection", this);
    comp->setShortcut(Qt::Key_C);
    comp->setCheckable(true);
    comp->setChecked(false);
    comp->setToolTip("Connected Component Selection (c)");
    selectionTypeGroup->addAction(comp);

    chooseSelectionType->addActions(selectionTypeGroup->actions());
    chooseSelectionType->setIconSize(QSize(40,40));
    vBoxTypeSel->addWidget(chooseSelectionType);


    QFrame* line1 = new QFrame();
    line1->setFrameShape(QFrame::VLine);
    line1->setFrameShadow(QFrame::Sunken);
    contentLayout->addWidget(line1,1);


    //***************************************************
    //                  stroke ToolBar
    //***************************************************
    QVBoxLayout* vBoxSelMan = new QVBoxLayout();
    contentLayout->addLayout(vBoxSelMan);

    QLabel* labelStroke = new QLabel("Edit");
    labelStroke->setFont(font);
    labelStroke->setAlignment(Qt::AlignCenter);
    vBoxSelMan->addWidget(labelStroke,1);

    controlStroke = new QToolBar;

    QActionGroup* strokeControllersBox = new QActionGroup(this);

    drawStroke = new QAction( QIcon("./GUI/icons/refine.png") , "refine selections boundaries", this);
    drawStroke->setCheckable(true);
    drawStroke->setChecked(false);
    drawStroke->setStatusTip("Draw Strokes");
    connect(drawStroke, SIGNAL(triggered()), this, SLOT(setSelectionManagementToZero0()));
    controlStroke->addAction(drawStroke);
    strokeControllersBox->addAction(drawStroke);

    eraseStroke = new QAction( QIcon("./GUI/icons/eraser.png") , "erase selections", this);
    eraseStroke->setCheckable(true);
    eraseStroke->setChecked(false);
    eraseStroke->setStatusTip("Erase strokes parts");
    connect(eraseStroke, SIGNAL(triggered()), this, SLOT(setSelectionManagementToZero1()));
    controlStroke->addAction(eraseStroke);
    strokeControllersBox->addAction(eraseStroke);
    strokeControllersBox->setExclusive(false);

    deleteSelection = new QAction( QIcon("./GUI/icons/trash2.png") , "delete a selection", this);
    deleteSelection->setCheckable(true);
    deleteSelection->setChecked(false);
    deleteSelection->setStatusTip("delete a stroke");
    connect(deleteSelection, SIGNAL(triggered()), this, SLOT(setSelectionManagementToZero2()));
    controlStroke->addAction(deleteSelection);
    strokeControllersBox->addAction(deleteSelection);

    deleteAll = new QAction( QIcon("./GUI/icons/deleteAll.png") , "delete all selections", this);
    deleteAll->setCheckable(false);
    deleteAll->setChecked(false);
    deleteAll->setStatusTip("reverse selection");
    connect(deleteAll, SIGNAL(triggered()), _viewer, SLOT(deleteAllSelections()));
    connect(deleteAll, SIGNAL(triggered()), this, SLOT(setSelectionManagementToZero3()));
    controlStroke->addAction(deleteAll);

    connect( _viewer, SIGNAL(resetStrokeToDraw()), this, SLOT(resetStrokeToDraw()));

    vBoxSelMan->addWidget(controlStroke);
    controlStroke->setDisabled(true);
    controlStroke->setIconSize(QSize(32,32));

    QFrame* line2 = new QFrame();
    line2->setFrameShape(QFrame::VLine);
    line2->setFrameShadow(QFrame::Sunken);
    contentLayout->addWidget(line2,1);

    //***************************************************
    //                  Expansion ToolBar
    //***************************************************
    QVBoxLayout* vBoxSim = new QVBoxLayout();
    contentLayout->addLayout(vBoxSim);

    QLabel* labelSelection = new QLabel("Expansion");
    labelSelection->setFont(font);
    labelSelection->setAlignment(Qt::AlignCenter);
    QHBoxLayout* hBox = new QHBoxLayout();

    vBoxSim->addWidget(labelSelection);
    vBoxSim->addLayout(hBox);

    QToolBar* controlSelection = new QToolBar;

    expansion = new QAction( QIcon("./GUI/icons/expansion2.png") , "Find Similar Selections", this);
    expansion->setShortcut(Qt::Key_S);
    expansion->setCheckable(false);
    expansion->setDisabled(false);
    expansion->setStatusTip("Find Similar Selections");
    connect(expansion, SIGNAL(triggered()), _viewer, SLOT(findSimilarSelections()));

    controlSelection->setIconSize(QSize(40,40));
    controlSelection->addAction(expansion);
    hBox->addWidget(controlSelection);

    similarityThreshold = new QSlider(Qt::Horizontal);
    similarityThreshold->setRange(0, 30);
    similarityThreshold->setValue(1);
    similarityThreshold->setMinimumWidth( 100 );
    similarityThreshold->setVisible(true);

    similarityValueLabel = new QLabel("");
    similarityValueLabel->setVisible(true);
    similarityValueLabel->setText(QString::number(similarityThreshold->value()));
    connect(similarityThreshold, SIGNAL(valueChanged(int)), this, SLOT(changeSimilarityValueLabel(int)));
    connect(similarityThreshold, SIGNAL(valueChanged(int)), _viewer, SLOT(thresholdMSChanged(int)));
    hBox->addWidget(similarityThreshold);
    hBox->addWidget(similarityValueLabel);

    controlStroke->setDisabled(true);
    controlStroke->setDisabled(true);
    expansion->setDisabled(true);
    addSelections->setDisabled(true);
    similarityThreshold->setDisabled(true);
    reverseSelection->setDisabled(true);
    this->setWidget(contents);
}

void SelectionDock::resetStrokeToDraw(){
    select->setChecked(true);
}

void SelectionDock::changeSimilarityValueLabel(int v){
    similarityValueLabel->setText(QString::number(v));
}

void SelectionDock::setSelectionType(QAction* action){
    similarSearchEnable = false;

    if (action->toolTip() == "Patch Selection (p)"){
        Trace(1, " selectionDock patch clicked");
        _viewer->setSelectionType(PATCH);
         similarSearchEnable = true;
    }
    else if (action->toolTip() == "Part Selection (m)"){
        Trace(1, " selectionDock part clicked ");
        _viewer->setSelectionType(PART);
        similarSearchEnable = true;
    } else if (action->toolTip() == "Connected Component Selection (c)") {
         Trace(1, " selectionDock comp clicked");
         _viewer->setSelectionType(COMP);
         similarSearchEnable = true;
    } else
        Trace(1, "error");

    if (similarSearchEnable) {
        expansion->setDisabled(false);
        similarityThreshold->setVisible(true);
    } else {
        expansion->setDisabled(true);
        similarityThreshold->setVisible(false);
    }
}


void SelectionDock::activateStrokeMode(bool b)
{
    controlStroke->setDisabled(!b);
    controlStroke->setDisabled(!b);
    expansion->setDisabled(!b);
    addSelections->setDisabled(!b);
    reverseSelection->setDisabled(!b);
    similarityThreshold->setDisabled(!b);
}

void SelectionDock::setSelectionManagementToZero0(){
    Trace(1, " selManagement 0");
    bool _checked = drawStroke->isChecked();
    _viewer->setDrawStrokeMode(_checked);
    eraseStroke->setChecked(false);
    deleteSelection->setChecked(false);
    deleteAll->setChecked(false);
}

void SelectionDock::setSelectionManagementToZero1(){
    Trace(1, " selManagement 1");
    bool _checked = eraseStroke->isChecked();
    _viewer->setEraserMode(_checked);
    drawStroke->setChecked(false);
    deleteSelection->setChecked(false);
    deleteAll->setChecked(false);
}

void SelectionDock::setSelectionManagementToZero2(){
    Trace(1, " selManagement 2");
    bool _checked = deleteSelection->isChecked();
    _viewer->setDeleteSelectionMode(_checked);
    drawStroke->setChecked(false);
    eraseStroke->setChecked(false);
    deleteAll->setChecked(false);
}

void SelectionDock::setSelectionManagementToZero3(){
    Trace(1, " selManagement 3");
    drawStroke->setChecked(false);
    eraseStroke->setChecked(false);
    deleteSelection->setChecked(false);
}
