/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "UserStroke.h"
#include "GLUtilityMethods.h"
#include "Debug.h"
#include "MathTools.h"

UserStroke::UserStroke()
{
    startNewStroke = true;
    numberOfStrokes = 0;
    strokeColor[0] = 0.f;
    strokeColor[1] = 0.03f;
    strokeColor[2] = 0.75f;
}

void UserStroke::draw(std::vector<float> & color)
{
    glDisable(GL_DEPTH_TEST);

    glEnable (GL_LINE_SMOOTH);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_LIGHTING);
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

    glLineWidth(2.f);
    glColor3f(color[0], color[1], color[2]);

    for (unsigned int i = 0; i < points.size(); i ++)
    {

        if (points[i].newStroke )
            glBegin(GL_LINE_STRIP);
            glVertexVec3Df(points[i].p);
        if (i == points.size()-1)
            glEnd();
        else if ( points[i+1].newStroke)
            glEnd();
    }

    if (points.size() < 5)
    {
        for (unsigned int i = 0; i < points.size(); i++)
            BasicGL::drawSphere(points[i].p[0], points[i].p[1], points[i].p[2], 0.001, 15, 15);
    }
}

void UserStroke::addPoint(Vec3Df p, float r)
{
    strokePoints sp;
    if (startNewStroke)
    {
        sp.newStroke = true;
        numberOfStrokes ++;
        lastStroke.clear();
    } else
        sp.newStroke = false;

    sp.p = p;
    sp.r = r;

    points.push_back(sp);
    startNewStroke = false;
}

void UserStroke::lastStrokeDrawn()
{
    startNewStroke = true;
}

void UserStroke::deletePoint(Vec3Df p, float r)
{
    strokePoints sp;
    sp.p =p;
    sp.r =r;
    sp.newStroke = startNewStroke;
    BGpoints.push_back(sp);

    if (startNewStroke)
        pointsBeingDeleted.clear();

    startNewStroke = false;
    for ( int i = 0; i < (int) points.size(); i++)
    {
        if ( (points[i].p-p).getLength() < r )
        {
            if (i < (int) points.size()-2)
            {
                if ( !points[i+1].newStroke )
                    points[i+1].newStroke=true;
            }
            pointsBeingDeleted.push_back(points[i]);
            points.erase(points.begin()+i);

        }
    }
    // count the number of strokes
    numberOfStrokes = 0;
    for (unsigned int i = 0; i< points.size();i++)
    {
        if (points[i].newStroke)
            numberOfStrokes++;
    }
}

void UserStroke::deleteStrokeUnderPoint(Vec3Df point)
{
    if (! points.empty())
    {
        if (startNewStroke)
            pointsBeingDeleted.clear();

        startNewStroke = false;

        float min = FLT_MAX;
        int ind = -1;
        for (unsigned int i = 0; i < points.size(); i++)
        {
            if ( (points[i].p-point).getLength() < min )
            {
                ind = i;
                min = (points[i].p-point).getLength();
            }
        }

        bool stop = false;
        int i = ind;
        while (! (stop || i < 0) )
        {
            if (points[i].newStroke)
                stop = true;
            else
                i--;
        }
        pointsBeingDeleted.push_back(points[i]);
        points.erase(points.begin()+i);
        stop =false;
        while (! (stop || points.empty() || points.size() < (unsigned int) i) )
        {
            if ( !points[i].newStroke )
            {
                pointsBeingDeleted.push_back(points[i]);
                points.erase(points.begin()+i);
            } else
                stop = true;
        }
        numberOfStrokes--;
    }
}

void UserStroke::deleteAllPoints()
{
    lastStroke.clear();
    points.clear();
    numberOfStrokes = 0;
    pointsBeingDeleted.clear();
}

